# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc
import warnings

from evochi.v1 import evochi_pb2 as evochi_dot_v1_dot_evochi__pb2

GRPC_GENERATED_VERSION = "1.67.0"
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + f" but the generated code in evochi/v1/evochi_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
    )


class EvochiServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Subscribe = channel.unary_stream(
            "/evochi.v1.EvochiService/Subscribe",
            request_serializer=evochi_dot_v1_dot_evochi__pb2.SubscribeRequest.SerializeToString,
            response_deserializer=evochi_dot_v1_dot_evochi__pb2.SubscribeResponse.FromString,
            _registered_method=True,
        )
        self.Heartbeat = channel.unary_unary(
            "/evochi.v1.EvochiService/Heartbeat",
            request_serializer=evochi_dot_v1_dot_evochi__pb2.HeartbeatRequest.SerializeToString,
            response_deserializer=evochi_dot_v1_dot_evochi__pb2.HeartbeatResponse.FromString,
            _registered_method=True,
        )
        self.FinishEvaluation = channel.unary_unary(
            "/evochi.v1.EvochiService/FinishEvaluation",
            request_serializer=evochi_dot_v1_dot_evochi__pb2.FinishEvaluationRequest.SerializeToString,
            response_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishEvaluationResponse.FromString,
            _registered_method=True,
        )
        self.FinishOptimization = channel.unary_unary(
            "/evochi.v1.EvochiService/FinishOptimization",
            request_serializer=evochi_dot_v1_dot_evochi__pb2.FinishOptimizationRequest.SerializeToString,
            response_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishOptimizationResponse.FromString,
            _registered_method=True,
        )
        self.FinishInitialization = channel.unary_unary(
            "/evochi.v1.EvochiService/FinishInitialization",
            request_serializer=evochi_dot_v1_dot_evochi__pb2.FinishInitializationRequest.SerializeToString,
            response_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishInitializationResponse.FromString,
            _registered_method=True,
        )
        self.FinishShareState = channel.unary_unary(
            "/evochi.v1.EvochiService/FinishShareState",
            request_serializer=evochi_dot_v1_dot_evochi__pb2.FinishShareStateRequest.SerializeToString,
            response_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishShareStateResponse.FromString,
            _registered_method=True,
        )


class EvochiServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Subscribe(self, request, context):
        """join the work force and subscribe to events"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Heartbeat(self, request, context):
        """send heartbeat to the server to keep the connection alive"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FinishEvaluation(self, request, context):
        """finish the evaluation"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FinishOptimization(self, request, context):
        """finish the optimization"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FinishInitialization(self, request, context):
        """finish the initialization"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FinishShareState(self, request, context):
        """finish the state sharing"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_EvochiServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Subscribe": grpc.unary_stream_rpc_method_handler(
            servicer.Subscribe,
            request_deserializer=evochi_dot_v1_dot_evochi__pb2.SubscribeRequest.FromString,
            response_serializer=evochi_dot_v1_dot_evochi__pb2.SubscribeResponse.SerializeToString,
        ),
        "Heartbeat": grpc.unary_unary_rpc_method_handler(
            servicer.Heartbeat,
            request_deserializer=evochi_dot_v1_dot_evochi__pb2.HeartbeatRequest.FromString,
            response_serializer=evochi_dot_v1_dot_evochi__pb2.HeartbeatResponse.SerializeToString,
        ),
        "FinishEvaluation": grpc.unary_unary_rpc_method_handler(
            servicer.FinishEvaluation,
            request_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishEvaluationRequest.FromString,
            response_serializer=evochi_dot_v1_dot_evochi__pb2.FinishEvaluationResponse.SerializeToString,
        ),
        "FinishOptimization": grpc.unary_unary_rpc_method_handler(
            servicer.FinishOptimization,
            request_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishOptimizationRequest.FromString,
            response_serializer=evochi_dot_v1_dot_evochi__pb2.FinishOptimizationResponse.SerializeToString,
        ),
        "FinishInitialization": grpc.unary_unary_rpc_method_handler(
            servicer.FinishInitialization,
            request_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishInitializationRequest.FromString,
            response_serializer=evochi_dot_v1_dot_evochi__pb2.FinishInitializationResponse.SerializeToString,
        ),
        "FinishShareState": grpc.unary_unary_rpc_method_handler(
            servicer.FinishShareState,
            request_deserializer=evochi_dot_v1_dot_evochi__pb2.FinishShareStateRequest.FromString,
            response_serializer=evochi_dot_v1_dot_evochi__pb2.FinishShareStateResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("evochi.v1.EvochiService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers("evochi.v1.EvochiService", rpc_method_handlers)


# This class is part of an EXPERIMENTAL API.
class EvochiService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Subscribe(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/evochi.v1.EvochiService/Subscribe",
            evochi_dot_v1_dot_evochi__pb2.SubscribeRequest.SerializeToString,
            evochi_dot_v1_dot_evochi__pb2.SubscribeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def Heartbeat(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/evochi.v1.EvochiService/Heartbeat",
            evochi_dot_v1_dot_evochi__pb2.HeartbeatRequest.SerializeToString,
            evochi_dot_v1_dot_evochi__pb2.HeartbeatResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def FinishEvaluation(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/evochi.v1.EvochiService/FinishEvaluation",
            evochi_dot_v1_dot_evochi__pb2.FinishEvaluationRequest.SerializeToString,
            evochi_dot_v1_dot_evochi__pb2.FinishEvaluationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def FinishOptimization(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/evochi.v1.EvochiService/FinishOptimization",
            evochi_dot_v1_dot_evochi__pb2.FinishOptimizationRequest.SerializeToString,
            evochi_dot_v1_dot_evochi__pb2.FinishOptimizationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def FinishInitialization(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/evochi.v1.EvochiService/FinishInitialization",
            evochi_dot_v1_dot_evochi__pb2.FinishInitializationRequest.SerializeToString,
            evochi_dot_v1_dot_evochi__pb2.FinishInitializationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def FinishShareState(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/evochi.v1.EvochiService/FinishShareState",
            evochi_dot_v1_dot_evochi__pb2.FinishShareStateRequest.SerializeToString,
            evochi_dot_v1_dot_evochi__pb2.FinishShareStateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class EventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    EVENT_TYPE_UNSPECIFIED: _ClassVar[EventType]
    EVENT_TYPE_HELLO: _ClassVar[EventType]
    EVENT_TYPE_EVALUATE: _ClassVar[EventType]
    EVENT_TYPE_OPTIMIZE: _ClassVar[EventType]
    EVENT_TYPE_INITIALIZE: _ClassVar[EventType]
    EVENT_TYPE_SHARE_STATE: _ClassVar[EventType]
    EVENT_TYPE_STOP: _ClassVar[EventType]

EVENT_TYPE_UNSPECIFIED: EventType
EVENT_TYPE_HELLO: EventType
EVENT_TYPE_EVALUATE: EventType
EVENT_TYPE_OPTIMIZE: EventType
EVENT_TYPE_INITIALIZE: EventType
EVENT_TYPE_SHARE_STATE: EventType
EVENT_TYPE_STOP: EventType

class Slice(_message.Message):
    __slots__ = ("start", "end")
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    start: int
    end: int
    def __init__(self, start: _Optional[int] = ..., end: _Optional[int] = ...) -> None: ...

class Evaluation(_message.Message):
    __slots__ = ("slice", "rewards")
    SLICE_FIELD_NUMBER: _ClassVar[int]
    REWARDS_FIELD_NUMBER: _ClassVar[int]
    slice: Slice
    rewards: _containers.RepeatedScalarFieldContainer[float]
    def __init__(
        self, slice: _Optional[_Union[Slice, _Mapping]] = ..., rewards: _Optional[_Iterable[float]] = ...
    ) -> None: ...

class HelloEvent(_message.Message):
    __slots__ = ("id", "token", "state", "population_size", "heartbeat_interval", "max_epochs")
    ID_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    POPULATION_SIZE_FIELD_NUMBER: _ClassVar[int]
    HEARTBEAT_INTERVAL_FIELD_NUMBER: _ClassVar[int]
    MAX_EPOCHS_FIELD_NUMBER: _ClassVar[int]
    id: str
    token: str
    state: bytes
    population_size: int
    heartbeat_interval: int
    max_epochs: int
    def __init__(
        self,
        id: _Optional[str] = ...,
        token: _Optional[str] = ...,
        state: _Optional[bytes] = ...,
        population_size: _Optional[int] = ...,
        heartbeat_interval: _Optional[int] = ...,
        max_epochs: _Optional[int] = ...,
    ) -> None: ...

class EvaluateEvent(_message.Message):
    __slots__ = ("task_id", "epoch", "slices")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    EPOCH_FIELD_NUMBER: _ClassVar[int]
    SLICES_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    epoch: int
    slices: _containers.RepeatedCompositeFieldContainer[Slice]
    def __init__(
        self,
        task_id: _Optional[str] = ...,
        epoch: _Optional[int] = ...,
        slices: _Optional[_Iterable[_Union[Slice, _Mapping]]] = ...,
    ) -> None: ...

class OptimizeEvent(_message.Message):
    __slots__ = ("task_id", "epoch", "rewards")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    EPOCH_FIELD_NUMBER: _ClassVar[int]
    REWARDS_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    epoch: int
    rewards: _containers.RepeatedScalarFieldContainer[float]
    def __init__(
        self, task_id: _Optional[str] = ..., epoch: _Optional[int] = ..., rewards: _Optional[_Iterable[float]] = ...
    ) -> None: ...

class InitializeEvent(_message.Message):
    __slots__ = ("task_id",)
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    def __init__(self, task_id: _Optional[str] = ...) -> None: ...

class ShareStateEvent(_message.Message):
    __slots__ = ("task_id", "epoch")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    EPOCH_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    epoch: int
    def __init__(self, task_id: _Optional[str] = ..., epoch: _Optional[int] = ...) -> None: ...

class StopEvent(_message.Message):
    __slots__ = ("task_id",)
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    def __init__(self, task_id: _Optional[str] = ...) -> None: ...

class SubscribeRequest(_message.Message):
    __slots__ = ("cores",)
    CORES_FIELD_NUMBER: _ClassVar[int]
    cores: int
    def __init__(self, cores: _Optional[int] = ...) -> None: ...

class SubscribeResponse(_message.Message):
    __slots__ = ("type", "hello", "evaluate", "optimize", "initialize", "share_state", "stop")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    HELLO_FIELD_NUMBER: _ClassVar[int]
    EVALUATE_FIELD_NUMBER: _ClassVar[int]
    OPTIMIZE_FIELD_NUMBER: _ClassVar[int]
    INITIALIZE_FIELD_NUMBER: _ClassVar[int]
    SHARE_STATE_FIELD_NUMBER: _ClassVar[int]
    STOP_FIELD_NUMBER: _ClassVar[int]
    type: EventType
    hello: HelloEvent
    evaluate: EvaluateEvent
    optimize: OptimizeEvent
    initialize: InitializeEvent
    share_state: ShareStateEvent
    stop: StopEvent
    def __init__(
        self,
        type: _Optional[_Union[EventType, str]] = ...,
        hello: _Optional[_Union[HelloEvent, _Mapping]] = ...,
        evaluate: _Optional[_Union[EvaluateEvent, _Mapping]] = ...,
        optimize: _Optional[_Union[OptimizeEvent, _Mapping]] = ...,
        initialize: _Optional[_Union[InitializeEvent, _Mapping]] = ...,
        share_state: _Optional[_Union[ShareStateEvent, _Mapping]] = ...,
        stop: _Optional[_Union[StopEvent, _Mapping]] = ...,
    ) -> None: ...

class HeartbeatRequest(_message.Message):
    __slots__ = ("seq_id", "timestamp")
    SEQ_ID_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    seq_id: int
    timestamp: _timestamp_pb2.Timestamp
    def __init__(
        self, seq_id: _Optional[int] = ..., timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...
    ) -> None: ...

class HeartbeatResponse(_message.Message):
    __slots__ = ("ok",)
    OK_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    def __init__(self, ok: bool = ...) -> None: ...

class FinishEvaluationRequest(_message.Message):
    __slots__ = ("task_id", "evaluations")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    EVALUATIONS_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    evaluations: _containers.RepeatedCompositeFieldContainer[Evaluation]
    def __init__(
        self, task_id: _Optional[str] = ..., evaluations: _Optional[_Iterable[_Union[Evaluation, _Mapping]]] = ...
    ) -> None: ...

class FinishEvaluationResponse(_message.Message):
    __slots__ = ("ok",)
    OK_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    def __init__(self, ok: bool = ...) -> None: ...

class FinishOptimizationRequest(_message.Message):
    __slots__ = ("task_id",)
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    def __init__(self, task_id: _Optional[str] = ...) -> None: ...

class FinishOptimizationResponse(_message.Message):
    __slots__ = ("ok",)
    OK_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    def __init__(self, ok: bool = ...) -> None: ...

class FinishInitializationRequest(_message.Message):
    __slots__ = ("task_id", "state")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    state: bytes
    def __init__(self, task_id: _Optional[str] = ..., state: _Optional[bytes] = ...) -> None: ...

class FinishInitializationResponse(_message.Message):
    __slots__ = ("ok",)
    OK_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    def __init__(self, ok: bool = ...) -> None: ...

class FinishShareStateRequest(_message.Message):
    __slots__ = ("task_id", "state")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    state: bytes
    def __init__(self, task_id: _Optional[str] = ..., state: _Optional[bytes] = ...) -> None: ...

class FinishShareStateResponse(_message.Message):
    __slots__ = ("ok",)
    OK_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    def __init__(self, ok: bool = ...) -> None: ...



from __future__ import annotations

from typing import AsyncIterable, NamedTuple
from abc import ABC, abstractmethod

import asyncio
import logging
import pickle
from datetime import datetime

import grpc.aio as grpc
from google.protobuf import timestamp_pb2
from grpc import StatusCode
import zstandard as zstd

import evochi.v1.evochi_pb2 as v1
import evochi.v1.evochi_pb2_grpc as client


class Eval(NamedTuple):
    """The evaluated slice and its rewards.

    Note that the len(rewards) must be equal to the length of the slice, which
    is `slice.stop - slice.start`.
    """

    slice: slice
    rewards: list[float]

    @staticmethod
    def from_flat(slices: list[slice], rewards: list[float]) -> list[Eval]:
        """Constructs a list of Eval objects from a list of slices and a list of
        rewards."""
        total_width = sum(sl.stop - sl.start for sl in slices)
        assert total_width == len(rewards), "Number of rewards must match the total width of the slices"
        evals: list[Eval] = []
        offset = 0
        for sl in slices:
            evals.append(
                Eval(
                    slice=slice(sl.start, sl.stop),
                    rewards=rewards[offset : offset + (sl.stop - sl.start)],
                )
            )
            offset += sl.stop - sl.start
        return evals


class Worker[S](ABC):
    def __init__(self, channel: grpc.Channel, cores: int) -> None:
        """Initializes the client to interact with the server via the given channel.

        Args:
            channel: The gRPC channel to use for communication.
            cores: The maximum number of cores to use for evaluation.
        """
        self._channel = channel
        self._cores = cores
        self._stub = client.EvochiServiceStub(channel)
        self._heartbeat_seq_id = 0
        self._closed = False
        self._token: str | None = None
        self._pop_size: int | None = None
        self._max_epochs: int | None = None
        self._heartbeat_interval: int = 0
        self._current_state: S | None = None

    @property
    def cores(self) -> int:
        """Returns the number of cores the worker is using."""
        return self._cores

    @property
    def population_size(self) -> int:
        if self._pop_size is None:
            raise RuntimeError("Worker has not been initialized yet")
        return self._pop_size

    @property
    def max_epochs(self) -> int:
        if self._max_epochs is None:
            raise RuntimeError("Worker has not been initialized yet")
        return self._max_epochs

    @property
    def state(self) -> S:
        if self._current_state is None:
            raise RuntimeError("Worker has not been initialized yet")
        return self._current_state

    @abstractmethod
    def initialize(self) -> S:
        """Lifecycle hook that is called for the first worker to initialize the
        state."""

    @abstractmethod
    def evaluate(self, epoch: int, slices: list[slice]) -> list[Eval]:
        """Lifecycle hook that is called when the worker should perform an
        evaluation step."""

    @abstractmethod
    def optimize(self, epoch: int, rewards: list[float]) -> S:
        """Lifecycle hook that is called when the worker should perform an
        optimization step."""

    def on_stop(self, cancel: bool) -> None:
        """Hook that is called when the worker is requested to stop.

        Args:
            cancel: Whether the worker's connection has been cancelled.
        """

    def on_state_change(self, state: S) -> None:
        """Hook that is called when the worker's state changes."""

    async def close(self) -> None:
        """Closes the client's channel."""
        if self._closed:
            raise RuntimeError("Worker is already closed")
        logging.debug("Closing worker")
        self._closed = True
        await self._channel.close()

    async def start(self) -> None:
        logging.debug("Starting worker with %d cores, waiting to be ready...", self._cores)
        await self._channel.channel_ready()
        logging.debug("Worker ready, starting...")
        await self._handle_events()

    async def _keep_alive(self) -> None:
        """Sends a heartbeat to the server periodically."""
        while not self._closed:
            self._heartbeat_seq_id += 1
            timestamp = timestamp_pb2.Timestamp()
            timestamp.FromDatetime(datetime.now())
            await self._heartbeat(
                v1.HeartbeatRequest(
                    seq_id=self._heartbeat_seq_id,
                    timestamp=timestamp,
                )
            )
            await asyncio.sleep(self._heartbeat_interval)

    async def _handle_events(self) -> None:
        """Event loop that listens for events from the server and dispatches them to the appropriate handlers."""
        iterator = self._subscribe(v1.SubscribeRequest(cores=self._cores))
        try:
            async for event in iterator:
                match event.type:
                    case v1.EVENT_TYPE_HELLO:
                        self._handle_hello_event(event.hello)
                    case v1.EVENT_TYPE_INITIALIZE:
                        await self._handle_init_event(event.initialize)
                    case v1.EVENT_TYPE_EVALUATE:
                        await self._handle_eval_event(event.evaluate)
                    case v1.EVENT_TYPE_SHARE_STATE:
                        await self._handle_share_state_event(event.share_state)
                    case v1.EVENT_TYPE_OPTIMIZE:
                        await self._handle_optimize_event(event.optimize)
                    case v1.EVENT_TYPE_STOP:
                        await self._handle_stop_event(event.stop)
                    case _:
                        logging.warning("Received unknown event type %s", event.type)
        except grpc.AioRpcError as e:
            if e.code() == StatusCode.CANCELLED:
                logging.debug("Worker cancelled")
                self.on_stop(True)
            else:
                raise e

    def _handle_hello_event(self, event: v1.HelloEvent) -> None:
        logging.debug(
            "Received hello event with id %s and token %s",
            event.id,
            event.token,
        )
        self._token = event.token
        self._heartbeat_interval = event.heartbeat_interval
        self._pop_size = event.population_size
        self._max_epochs = event.max_epochs

        self._current_state = self._decompress_state(event.state) if event.state else None
        if self._current_state is not None:
            self.on_state_change(self._current_state)

        asyncio.create_task(self._keep_alive())

    async def _handle_init_event(self, event: v1.InitializeEvent) -> None:
        logging.debug("Received init event with task id %s", event.task_id)
        state = self.initialize()

        self._current_state = state
        self.on_state_change(self._current_state)

        await self._finish_initialization(
            v1.FinishInitializationRequest(
                task_id=event.task_id,
                state=self._compressed_state(),
            )
        )

    async def _handle_eval_event(self, event: v1.EvaluateEvent) -> None:
        logging.debug("Received eval event with task id %s", event.task_id)
        evals = self.evaluate(event.epoch, [slice(sl.start, sl.end) for sl in event.slices])
        await self._finish_evaluation(
            v1.FinishEvaluationRequest(
                task_id=event.task_id,
                evaluations=[
                    v1.Evaluation(
                        slice=v1.Slice(start=e.slice.start, end=e.slice.stop),
                        rewards=e.rewards,
                    )
                    for e in evals
                ],
            )
        )

    async def _handle_optimize_event(self, event: v1.OptimizeEvent) -> None:
        logging.debug("Received optimize event with task id %s", event.task_id)
        optimized = self.optimize(event.epoch, list(event.rewards))

        self._current_state = optimized
        self.on_state_change(optimized)

        await self._finish_optimization(v1.FinishOptimizationRequest(task_id=event.task_id))

    async def _handle_share_state_event(self, event: v1.ShareStateEvent) -> None:
        logging.debug("Received share state event with task id %s", event.task_id)
        assert self._current_state is not None
        await self._finish_share_state(
            v1.FinishShareStateRequest(
                task_id=event.task_id,
                state=self._compressed_state(),
            )
        )

    async def _handle_stop_event(self, event: v1.StopEvent) -> None:
        logging.debug("Received stop event with task id %s", event.task_id)
        self.on_stop(False)

    def _subscribe(self, request: v1.SubscribeRequest) -> AsyncIterable[v1.SubscribeResponse]:
        logging.debug("Subscribing to events")
        return self._stub.Subscribe(request)

    async def _heartbeat(self, request: v1.HeartbeatRequest) -> v1.HeartbeatResponse:
        logging.debug("Sending heartbeat")
        return await self._stub.Heartbeat(request, metadata=self._metadata())

    async def _finish_evaluation(self, request: v1.FinishEvaluationRequest) -> v1.FinishEvaluationResponse:
        logging.debug("Sending finish evaluation")
        return await self._stub.FinishEvaluation(request, metadata=self._metadata())

    async def _finish_optimization(self, request: v1.FinishOptimizationRequest) -> v1.FinishOptimizationResponse:
        logging.debug("Sending finish optimization")
        return await self._stub.FinishOptimization(request, metadata=self._metadata())

    async def _finish_initialization(self, request: v1.FinishInitializationRequest) -> v1.FinishInitializationResponse:
        logging.debug("Sending finish initialization")
        return await self._stub.FinishInitialization(request, metadata=self._metadata())

    async def _finish_share_state(self, request: v1.FinishShareStateRequest) -> v1.FinishShareStateResponse:
        logging.debug("Sending finish share state")
        return await self._stub.FinishShareState(request, metadata=self._metadata())

    def _compressed_state(self) -> bytes:
        """Returns the current state compressed using zstandard."""
        return zstd.compress(pickle.dumps(self._current_state))

    @staticmethod
    def _decompress_state(state: bytes) -> S:
        """Returns the current state decompressed using zstandard."""
        return pickle.loads(zstd.decompress(state))

    def _metadata(self) -> list[tuple[str, str]]:
        """Returns the metadata to use for requests to the server."""
        if self._token is None:
            raise RuntimeError("Client does not have a token yet")
        return [("authorization", f"Bearer {self._token}")]


__all__ = ["Worker", "Eval"]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: evochi/v1/evochi.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(_runtime_version.Domain.PUBLIC, 5, 27, 2, "", "evochi/v1/evochi.proto")
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x16\x65vochi/v1/evochi.proto\x12\tevochi.v1\x1a\x1fgoogle/protobuf/timestamp.proto"#\n\x05Slice\x12\r\n\x05start\x18\x01 \x01(\x05\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x05">\n\nEvaluation\x12\x1f\n\x05slice\x18\x01 \x01(\x0b\x32\x10.evochi.v1.Slice\x12\x0f\n\x07rewards\x18\x02 \x03(\x02"\x8e\x01\n\nHelloEvent\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05token\x18\x02 \x01(\t\x12\x12\n\x05state\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x12\x17\n\x0fpopulation_size\x18\x04 \x01(\x05\x12\x1a\n\x12heartbeat_interval\x18\x05 \x01(\x05\x12\x12\n\nmax_epochs\x18\x06 \x01(\x05\x42\x08\n\x06_state"Q\n\rEvaluateEvent\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\r\n\x05\x65poch\x18\x02 \x01(\x05\x12 \n\x06slices\x18\x03 \x03(\x0b\x32\x10.evochi.v1.Slice"@\n\rOptimizeEvent\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\r\n\x05\x65poch\x18\x02 \x01(\x05\x12\x0f\n\x07rewards\x18\x03 \x03(\x02""\n\x0fInitializeEvent\x12\x0f\n\x07task_id\x18\x01 \x01(\t"1\n\x0fShareStateEvent\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\r\n\x05\x65poch\x18\x02 \x01(\x05"\x1c\n\tStopEvent\x12\x0f\n\x07task_id\x18\x01 \x01(\t"!\n\x10SubscribeRequest\x12\r\n\x05\x63ores\x18\x01 \x01(\x05"\xcf\x02\n\x11SubscribeResponse\x12"\n\x04type\x18\x01 \x01(\x0e\x32\x14.evochi.v1.EventType\x12&\n\x05hello\x18\x02 \x01(\x0b\x32\x15.evochi.v1.HelloEventH\x00\x12,\n\x08\x65valuate\x18\x03 \x01(\x0b\x32\x18.evochi.v1.EvaluateEventH\x00\x12,\n\x08optimize\x18\x04 \x01(\x0b\x32\x18.evochi.v1.OptimizeEventH\x00\x12\x30\n\ninitialize\x18\x05 \x01(\x0b\x32\x1a.evochi.v1.InitializeEventH\x00\x12\x31\n\x0bshare_state\x18\x06 \x01(\x0b\x32\x1a.evochi.v1.ShareStateEventH\x00\x12$\n\x04stop\x18\x07 \x01(\x0b\x32\x14.evochi.v1.StopEventH\x00\x42\x07\n\x05\x65vent"Q\n\x10HeartbeatRequest\x12\x0e\n\x06seq_id\x18\x01 \x01(\x05\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"\x1f\n\x11HeartbeatResponse\x12\n\n\x02ok\x18\x01 \x01(\x08"V\n\x17\x46inishEvaluationRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12*\n\x0b\x65valuations\x18\x02 \x03(\x0b\x32\x15.evochi.v1.Evaluation"&\n\x18\x46inishEvaluationResponse\x12\n\n\x02ok\x18\x01 \x01(\x08",\n\x19\x46inishOptimizationRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t"(\n\x1a\x46inishOptimizationResponse\x12\n\n\x02ok\x18\x01 \x01(\x08"=\n\x1b\x46inishInitializationRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\r\n\x05state\x18\x02 \x01(\x0c"*\n\x1c\x46inishInitializationResponse\x12\n\n\x02ok\x18\x01 \x01(\x08"9\n\x17\x46inishShareStateRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\r\n\x05state\x18\x02 \x01(\x0c"&\n\x18\x46inishShareStateResponse\x12\n\n\x02ok\x18\x01 \x01(\x08*\xbb\x01\n\tEventType\x12\x1a\n\x16\x45VENT_TYPE_UNSPECIFIED\x10\x00\x12\x14\n\x10\x45VENT_TYPE_HELLO\x10\x01\x12\x17\n\x13\x45VENT_TYPE_EVALUATE\x10\x03\x12\x17\n\x13\x45VENT_TYPE_OPTIMIZE\x10\x04\x12\x19\n\x15\x45VENT_TYPE_INITIALIZE\x10\x05\x12\x1a\n\x16\x45VENT_TYPE_SHARE_STATE\x10\x06\x12\x13\n\x0f\x45VENT_TYPE_STOP\x10\x07\x32\xb3\x04\n\rEvochiService\x12J\n\tSubscribe\x12\x1b.evochi.v1.SubscribeRequest\x1a\x1c.evochi.v1.SubscribeResponse"\x00\x30\x01\x12H\n\tHeartbeat\x12\x1b.evochi.v1.HeartbeatRequest\x1a\x1c.evochi.v1.HeartbeatResponse"\x00\x12]\n\x10\x46inishEvaluation\x12".evochi.v1.FinishEvaluationRequest\x1a#.evochi.v1.FinishEvaluationResponse"\x00\x12\x63\n\x12\x46inishOptimization\x12$.evochi.v1.FinishOptimizationRequest\x1a%.evochi.v1.FinishOptimizationResponse"\x00\x12i\n\x14\x46inishInitialization\x12&.evochi.v1.FinishInitializationRequest\x1a\'.evochi.v1.FinishInitializationResponse"\x00\x12]\n\x10\x46inishShareState\x12".evochi.v1.FinishShareStateRequest\x1a#.evochi.v1.FinishShareStateResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "evochi.v1.evochi_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_EVENTTYPE"]._serialized_start = 1494
    _globals["_EVENTTYPE"]._serialized_end = 1681
    _globals["_SLICE"]._serialized_start = 70
    _globals["_SLICE"]._serialized_end = 105
    _globals["_EVALUATION"]._serialized_start = 107
    _globals["_EVALUATION"]._serialized_end = 169
    _globals["_HELLOEVENT"]._serialized_start = 172
    _globals["_HELLOEVENT"]._serialized_end = 314
    _globals["_EVALUATEEVENT"]._serialized_start = 316
    _globals["_EVALUATEEVENT"]._serialized_end = 397
    _globals["_OPTIMIZEEVENT"]._serialized_start = 399
    _globals["_OPTIMIZEEVENT"]._serialized_end = 463
    _globals["_INITIALIZEEVENT"]._serialized_start = 465
    _globals["_INITIALIZEEVENT"]._serialized_end = 499
    _globals["_SHARESTATEEVENT"]._serialized_start = 501
    _globals["_SHARESTATEEVENT"]._serialized_end = 550
    _globals["_STOPEVENT"]._serialized_start = 552
    _globals["_STOPEVENT"]._serialized_end = 580
    _globals["_SUBSCRIBEREQUEST"]._serialized_start = 582
    _globals["_SUBSCRIBEREQUEST"]._serialized_end = 615
    _globals["_SUBSCRIBERESPONSE"]._serialized_start = 618
    _globals["_SUBSCRIBERESPONSE"]._serialized_end = 953
    _globals["_HEARTBEATREQUEST"]._serialized_start = 955
    _globals["_HEARTBEATREQUEST"]._serialized_end = 1036
    _globals["_HEARTBEATRESPONSE"]._serialized_start = 1038
    _globals["_HEARTBEATRESPONSE"]._serialized_end = 1069
    _globals["_FINISHEVALUATIONREQUEST"]._serialized_start = 1071
    _globals["_FINISHEVALUATIONREQUEST"]._serialized_end = 1157
    _globals["_FINISHEVALUATIONRESPONSE"]._serialized_start = 1159
    _globals["_FINISHEVALUATIONRESPONSE"]._serialized_end = 1197
    _globals["_FINISHOPTIMIZATIONREQUEST"]._serialized_start = 1199
    _globals["_FINISHOPTIMIZATIONREQUEST"]._serialized_end = 1243
    _globals["_FINISHOPTIMIZATIONRESPONSE"]._serialized_start = 1245
    _globals["_FINISHOPTIMIZATIONRESPONSE"]._serialized_end = 1285
    _globals["_FINISHINITIALIZATIONREQUEST"]._serialized_start = 1287
    _globals["_FINISHINITIALIZATIONREQUEST"]._serialized_end = 1348
    _globals["_FINISHINITIALIZATIONRESPONSE"]._serialized_start = 1350
    _globals["_FINISHINITIALIZATIONRESPONSE"]._serialized_end = 1392
    _globals["_FINISHSHARESTATEREQUEST"]._serialized_start = 1394
    _globals["_FINISHSHARESTATEREQUEST"]._serialized_end = 1451
    _globals["_FINISHSHARESTATERESPONSE"]._serialized_start = 1453
    _globals["_FINISHSHARESTATERESPONSE"]._serialized_end = 1491
    _globals["_EVOCHISERVICE"]._serialized_start = 1684
    _globals["_EVOCHISERVICE"]._serialized_end = 2247
# @@protoc_insertion_point(module_scope)
